import datetime
import os
import time
from datetime import timedelta
from functools import lru_cache

# ---------------------------------------------------------------------------


@lru_cache(maxsize=1)
def has_extra_cpu():
    try:
        import psutil
    except ImportError:
        return False
    return True


@lru_cache(maxsize=1)
def has_extra_gpu():
    try:
        import GPUtil
    except ImportError:
        return False
    return True


def get_cpu_info():
    if not has_extra_cpu():
        return None

    import psutil

    meminfo = psutil.virtual_memory()
    GB_div = 1024 ** 3

    info = (
        "```\n"
        + "\n".join(
            [
                f"Uptime:  {timedelta(seconds=int(time.time() - psutil.boot_time()))}",
                f"CPUs:    {psutil.cpu_count()}",
                f"RAM:     {meminfo.total / GB_div:.1f} GB",
                "",
                "Load:    1min: {0[0]:.1f}%, 5min: {0[1]:.1f}%, 15min: {0[2]:.1f}%".format(
                    [x / psutil.cpu_count() * 100 for x in psutil.getloadavg()]
                ),
                f"Memory:  {(meminfo.used / meminfo.total) * 100:.1f}% [used: {meminfo.used / GB_div:.1f} / {meminfo.total / GB_div:.1f} GB] [available: {meminfo.available  / GB_div:.1f} GB]",
            ]
        )
        + "\n```"
    )

    return info


def get_disk_info():
    if not has_extra_cpu():
        return None

    import psutil
    from psutil._common import bytes2human

    info = ""

    disks = [
        disk
        for disk in psutil.disk_partitions(all=False)
        if "loop" not in disk.device and not disk.mountpoint.startswith("/boot")
    ]

    header = ("Device", "Mount", "Use", "Total", "Used", "Free")
    rows = list()
    for disk in disks:
        usage = psutil.disk_usage(disk.mountpoint)
        rows.append(
            (
                disk.device,
                disk.mountpoint,
                f"{usage.percent:.1f} %",
                bytes2human(usage.total),
                bytes2human(usage.used),
                bytes2human(usage.free),
                # disk.fstype,
            )
        )

    lengths = [
        max(len(row[field_idx]) for row in [header] + rows)
        for field_idx in range(len(rows[0]))
    ]

    info = (
        "```\n"
        + "\n".join(
            # header
            [
                # "| " +
                " | ".join(
                    [
                        f"{field:{field_len}s}"
                        for field, field_len in zip(header, lengths)
                    ]
                )
                # + " |"
            ]
            # separator
            + [
                # "| " +
                " | ".join(["-" * field_len for field_len in lengths])
                # + " |"
            ]
            # rows
            + [
                # "| " +
                " | ".join(
                    # text fields
                    [
                        f"{field:<{field_len}s}"
                        for field, field_len in list(zip(row, lengths))[:2]
                    ]
                    # values/number
                    + [
                        f"{field:>{field_len}s}"
                        for field, field_len in list(zip(row, lengths))[2:]
                    ]
                )
                # + " |"
                for row in rows
            ]
        )
        + "\n```"
    )

    return info


def get_gpu_info():
    if not has_extra_gpu():
        return None

    import GPUtil

    rows = list()
    fields = ["ID", "Util", "Mem", "Temp", "Memory (Used)"]  # , "Name"]
    rows.append(fields)

    for gpu in GPUtil.getGPUs():
        fields = [
            f"{gpu.id}",
            f"{gpu.load * 100:.0f} %",
            f"{gpu.memoryUtil * 100:.1f} %",
            f"{gpu.temperature:.1f} °C",
            f"{int(gpu.memoryUsed)} / {int(gpu.memoryTotal)} MB",
            # f"{gpu.name}",
        ]
        rows.append(fields)

    lengths = [
        max(len(row[field_idx]) for row in rows) for field_idx in range(len(rows[0]))
    ]

    info = (
        "```\n"
        + "\n".join(
            # header
            [
                # "| " +
                " | ".join(
                    [
                        f"{field:{field_len}s}"
                        for field, field_len in zip(rows[0], lengths)
                    ]
                )
                # + " |"
            ]
            # separator
            + [
                # "| " +
                " | ".join(["-" * field_len for field_len in lengths])
                # + " |"
            ]
            # rows
            + [
                # "| " +
                " | ".join(
                    [f"{field:>{field_len}s}" for field, field_len in zip(row, lengths)]
                )
                # + " |"
                for row in rows[1:]
            ]
        )
        + "\n```"
    )

    return info


def get_info_message(with_cpu=True, with_gpu=True):
    message = f"**Status of `{os.uname().nodename}`**\n"
    message += f"Date: `{datetime.datetime.now()}`\n\n"

    if with_cpu and has_extra_cpu():
        message += "System information:"
        ret = get_cpu_info()
        if ret is not None:
            message += "\n" + ret + "\n"
        else:
            message += " N/A\n"

        message += "Disk information:"
        ret = get_disk_info()
        if ret is not None:
            message += "\n" + ret + "\n"
        else:
            message += " N/A\n"

    if with_gpu and has_extra_gpu():
        message += "GPU information:"
        ret = get_gpu_info()
        if ret is not None:
            message += "\n" + ret + "\n"
        else:
            message += " N/A\n"

    return message


# ---------------------------------------------------------------------------



import logging
import os
import discord

LOGGER = logging.getLogger(__name__)

# ---------------------------------------------------------------------------


class AbstractSingleActionClient(discord.Client):
    async def do_work(self):
        raise NotImplementedError()

    async def on_ready(self):
        LOGGER.info(f"Logged on as {self.user}")
        LOGGER.debug(f"name: {self.user.name}, id: {self.user.id}")

        await self.do_work()

        await self.close()


# ---------------------------------------------------------------------------


def send_message(token, channel_id, message):
    class SendMessageClient(AbstractSingleActionClient):
        async def do_work(self):
            channel = self.get_channel(channel_id)
            LOGGER.info(f"Channel: {channel} {type(channel)} {repr(channel)}")

            result = await channel.send(message)
            LOGGER.debug(f"MSG result: {result} {type(result)} {repr(result)}")

    client = SendMessageClient()
    client.run(token)


def send_file(token, channel_id, message, filename):
    name = os.path.basename(filename)  # only show last part of path
    # wrap file for discord
    dfile = discord.File(filename, filename=name)

    class SendFileMessageClient(AbstractSingleActionClient):
        async def do_work(self):
            channel = self.get_channel(channel_id)
            LOGGER.info(f"Channel: {channel} {type(channel)} {repr(channel)}")

            # attach file to message
            result = await channel.send(message, file=dfile)
            LOGGER.debug(f"MSG result: {result} {type(result)} {repr(result)}")

    client = SendFileMessageClient()
    client.run(token)


# ---------------------------------------------------------------------------



from setuptools import setup


def load_content(filename):
    with open(filename, "r", encoding="utf-8") as fp:
        return fp.read()


setup(
    name="discord-notifier-bot",
    version="0.2.0",
    license="MIT License",
    author="Erik Körner",
    author_email="koerner@informatik.uni-leipzig.de",
    description="A cli bot to send simple messages to a Discord channel.",
    long_description=load_content("README.rst"),
    long_description_content_type="text/x-rst",
    url="https://github.com/Querela/discord-notifier-bot",
    keywords=["discord", "bot", "notifier", "cli"],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Utilities",
    ],
    packages=["discord_notifier_bot"],
    python_requires=">=3.6",
    install_requires=["discord.py"],
    extras_require={"cpu": ["psutil"], "gpu": ["gputil"],},
    entry_points={
        "console_scripts": [
            "dbot-run = discord_notifier_bot.cli:main",
            "dbot-message = discord_notifier_bot.cli:main_message",
            "dbot-file = discord_notifier_bot.cli:main_file",
            "dbot-info = discord_notifier_bot.cli:main_info",
        ]
    },
)



import argparse
import configparser
import logging
import os
import pathlib
import sys

from discord_notifier_bot.bot import send_message as bot_send_message
from discord_notifier_bot.bot import send_file as bot_send_file
from discord_notifier_bot.sysinfo import has_extra_cpu
from discord_notifier_bot.sysinfo import has_extra_gpu
from discord_notifier_bot.sysinfo import get_info_message

LOGGER = logging.getLogger(__name__)

# ---------------------------------------------------------------------------


CONFIG_SECTION_NAME = "discord-bot"

CONFIG_PATHS = [
    pathlib.Path.home() / ".dbot.conf",
    pathlib.Path.home() / "dbot.conf",
    pathlib.Path(".") / ".dbot.conf",
    pathlib.Path(".") / "dbot.conf",
    pathlib.Path("/etc/dbot.conf"),
]


def find_config_file():
    for filename in CONFIG_PATHS:
        if filename.exists():
            LOGGER.info(f"Found config file: {filename}")
            return filename

    LOGGER.error("Found no configuration file in search path!")
    return None


def load_config_file(filename):
    config = configparser.ConfigParser()

    LOGGER.debug(f"Try loading configurations from {filename}")

    try:
        config.read(filename)

        if CONFIG_SECTION_NAME not in config:
            LOGGER.error(f"Missing configuration section header: {CONFIG_SECTION_NAME}")
            return None

        configs = config[CONFIG_SECTION_NAME]

        return {
            "token": configs["token"].strip('"'),
            "channel": int(configs["channel"]),
        }
    except KeyError as ex:
        LOGGER.error(f"Missing configuration key! >>{ex.args[0]}<<")
    except:
        LOGGER.exception("Loading configuration failed!")
    return None


def load_config(filename=None, **kwargs):
    configs = None

    if filename and os.path.isfile(filename):
        configs = load_config_file(filename)
        if configs is None:
            LOGGER.error("Loading given config file failed! Trying default ones ...")

    if configs is None:
        filename = find_config_file()
        if filename is not None:
            configs = load_config_file(filename)

    if configs is None:
        if "token" not in kwargs or "channel" not in kwargs:
            raise Exception("No configuration file found!")

    configs = {**configs, **kwargs}

    return configs


# ---------------------------------------------------------------------------


def send_message(bot_token, channel_id, message):
    LOGGER.info(f"Send message: {message} ...")
    bot_send_message(bot_token, channel_id, message)


def send_file(bot_token, channel_id, message, filename):
    if not os.path.isfile(filename):
        raise Exception(f"filename '{filename}' is not a file!")

    LOGGER.info(f"Send file: {filename} ...")
    bot_send_file(bot_token, channel_id, message, filename)


# ---------------------------------------------------------------------------


def parse_args(args=None):
    parser = argparse.ArgumentParser()

    actions = ("message", "file")
    if has_extra_cpu() or has_extra_gpu():
        actions += ("info",)

    parser.add_argument("action", choices=actions, help="Bot action")
    parser.add_argument("message", help="Message to send")
    parser.add_argument(
        "--type",
        type=str,
        nargs="?",
        const="",
        default=None,
        help="Markdown type for text messages, or default if as flag",
    )
    parser.add_argument(
        "-f",
        "--file",
        required=False,
        type=str,
        default=None,
        help="Optional file to attach to message",
    )
    parser.add_argument("-c", "--config", type=str, default=None, help="Config file")
    parser.add_argument(
        "-d", "--debug", action="store_true", help="Enable debug logging"
    )

    args = parser.parse_args(args)
    return args


# ---------------------------------------------------------------------------


def main(args=None):
    args = parse_args(args)

    if args.debug:
        # logging.basicConfig(format="* %(message)s", level=logging.INFO)
        logging.basicConfig(
            format="[%(levelname).1s] {%(name)s} %(message)s", level=logging.DEBUG
        )
        logging.getLogger("websockets.protocol").setLevel(logging.WARNING)
        logging.getLogger("websockets.client").setLevel(logging.WARNING)
        logging.getLogger("discord.client").setLevel(logging.WARNING)

    configs = load_config(filename=args.config)
    LOGGER.debug(f"Run bot with configs: {configs}")

    if args.message == "-":
        LOGGER.debug("Read message from STDIN ...")
        message = sys.stdin.read()
        LOGGER.debug(f"Read {len(message)} characters.")
    else:
        message = args.message

    if args.type is not None:
        LOGGER.info(f"Wrap message in markdown, type={args.type}")
        message = f"```{args.type}\n{message}\n```"

    if args.action == "message":
        send_message(configs["token"], configs["channel"], message)
    elif args.action == "file":
        send_file(configs["token"], configs["channel"], message, args.file)
    elif args.action == "info":
        # gather cpu/gpu infos and send message
        if message:
            message += "\n\n"
        message += get_info_message()
        send_message(configs["token"], configs["channel"], message)

    LOGGER.info("Done.")


# ---------------------------------------------------------------------------


def main_message():
    try:
        parser = argparse.ArgumentParser()
        parser.add_argument("message")
        parser.add_argument(
            "-c", "--config", type=str, default=None, help="Config file"
        )
        parser.add_argument(
            "--type",
            type=str,
            nargs="?",
            const="",
            default=None,
            help="Markdown type, or default if without value",
        )
        args = parser.parse_args()

        configs = load_config(filename=args.config)
        LOGGER.debug(f"Run bot with configs: {configs}")

        if args.message == "-":
            message = sys.stdin.read()
        else:
            message = args.message

        if args.type is not None:
            LOGGER.info(f"Wrap message in markdown, type={args.type}")
            message = f"```{args.type}\n{message}\n```"

        send_message(configs["token"], configs["channel"], message)
    except:
        sys.exit(1)


def main_file():
    try:
        parser = argparse.ArgumentParser()
        parser.add_argument("file")
        parser.add_argument("message")
        parser.add_argument(
            "-c", "--config", type=str, default=None, help="Config file"
        )
        args = parser.parse_args()

        configs = load_config(filename=args.config)
        LOGGER.debug(f"Run bot with configs: {configs}")

        send_file(configs["token"], configs["channel"], args.message, args.file)
    except:
        sys.exit(1)


def main_info():
    try:
        parser = argparse.ArgumentParser()
        parser.add_argument(
            "-c", "--config", type=str, default=None, help="Config file"
        )
        args = parser.parse_args()

        configs = load_config(filename=args.config)
        LOGGER.debug(f"Run bot with configs: {configs}")

        if not has_extra_cpu() and not has_extra_gpu():
            import warnings

            warnings.warn(
                "discord-notifier-bot has no extra dependencies installed to support the 'info' function!"
            )
            sys.exit(1)

        message = get_info_message()
        send_message(configs["token"], configs["channel"], message)
    except:
        sys.exit(1)


# ---------------------------------------------------------------------------
